from enum import Enum
import copy
import math

from sqc.src.Operand import Operand


# Enum of possible binary operations.
class BinOp(Enum):
    Add = 1
    Sub = 2
    Mult = 3
    Div = 4
    Pow = 5
    log = 6
    mod = 7


class Compare(Enum):
    Eq = 1
    NotEq = 2
    Lt = 3
    LtE = 4
    Gt = 5
    GtE = 6


class VectorBinOp(Enum):
    Add = 1
    Sub = 2
    MultByNum = 3
    DotProd = 4
    GetInd = 5
    Mult = 6
    TensMult = 7


# Enum of possible unary operations.
class UnOp(Enum):
    UAdd = 1
    USub = 2
    sin = 3
    cos = 4
    tg = 5
    ctg = 6
    exp = 7
    abs = 8
    arcsin = 9
    arccos = 10
    arctg = 11
    sinh = 12
    cosh = 13
    tgh = 14
    ctgh = 15
    arcctg = 16
    ln = 17
    log2 = 18
    log10 = 19
    arcsinh = 20
    arccosh = 21
    arctgh = 22


class VectorUnOp(Enum):
    USub = 1
    Transpose = 2


class VectorOp(Enum):
    Vector = 1


# Correspondence between custom and ast available binary operations
ast_bin_ops = {
    'Add': BinOp.Add,
    'Sub': BinOp.Sub,
    'Mult': BinOp.Mult,
    'Div': BinOp.Div,
    'Pow': BinOp.Pow,
    'log': BinOp.log,
    'mod': BinOp.mod,
    'vAdd': VectorBinOp.Add,
    'vSub': VectorBinOp.Sub,
    'vMulNum': VectorBinOp.MultByNum,
    'vDP': VectorBinOp.DotProd,
    'vGet': VectorBinOp.GetInd,
    'vMul': VectorBinOp.Mult,
    'vTMul': VectorBinOp.TensMult
}

ast_compare_ops = {
    'Eq': Compare.Eq,
    'NotEq': Compare.NotEq,
    'Lt': Compare.Lt,
    'LtE': Compare.LtE,
    'Gt': Compare.Gt,
    'GtE': Compare.GtE
}

functions = {
    UnOp.sin,
    UnOp.cos,
    UnOp.tg,
    UnOp.ctg,
    UnOp.exp,
    UnOp.abs,
    UnOp.sinh,
    UnOp.cosh,
    UnOp.tgh,
    UnOp.ctgh,
    UnOp.arcsin,
    UnOp.arccos,
    UnOp.arctg,
    UnOp.arcctg,
    UnOp.ln,
    UnOp.log2,
    UnOp.log10,
    BinOp.log,
    BinOp.mod,
    UnOp.arcsinh,
    UnOp.arccosh,
    UnOp.arctgh,
    VectorBinOp.Add,
    VectorBinOp.Sub,
    VectorBinOp.MultByNum,
    VectorBinOp.DotProd,
    VectorBinOp.TensMult,
    VectorBinOp.Mult,
    VectorUnOp.USub,
    VectorUnOp.Transpose
}


# Correspondence between custom and ast available unary operations
ast_un_ops = {
    'UAdd': UnOp.UAdd,
    'USub': UnOp.USub,
    'sin': UnOp.sin,
    'cos': UnOp.cos,
    'tg': UnOp.tg,
    'ctg': UnOp.ctg,
    'exp': UnOp.exp,
    'abs': UnOp.abs,
    'sinh': UnOp.sinh,
    'cosh': UnOp.cosh,
    'tgh': UnOp.tgh,
    'ctgh': UnOp.ctgh,
    'arcsin': UnOp.arcsin,
    'arccos': UnOp.arccos,
    'arctg': UnOp.arctg,
    'arcctg': UnOp.arcctg,
    'ln': UnOp.ln,
    'log2': UnOp.log2,
    'log10': UnOp.log10,
    'arcsinh': UnOp.arcsinh,
    'arccosh': UnOp.arccosh,
    'arctgh': UnOp.arctgh,
    'vUSub': VectorUnOp.USub,
    'vT': VectorUnOp.Transpose
}


oper_symbols = {
    BinOp.Add: '+',
    BinOp.Sub: '-',
    BinOp.Mult: '*',
    BinOp.Div: '/',
    BinOp.Pow: '**',
    UnOp.UAdd: '+',
    UnOp.USub: '-',
    UnOp.sin: 'sin',
    UnOp.cos: 'cos',
    UnOp.tg: 'tg',
    UnOp.ctg: 'ctg',
    UnOp.exp: 'exp',
    UnOp.abs: 'abs',
    UnOp.sinh: 'sinh',
    UnOp.cosh: 'cosh',
    UnOp.tgh: 'tgh',
    UnOp.ctgh: 'ctgh',
    UnOp.arcsin: 'arcsin',
    UnOp.arccos: 'arccos',
    UnOp.arctg: 'arctg',
    UnOp.arcctg: 'arcctg',
    UnOp.ln: 'ln',
    UnOp.log2: 'log2',
    UnOp.log10: 'log10',
    BinOp.log: 'log',
    BinOp.mod: 'mod',
    UnOp.arcsinh: 'arcsinh',
    UnOp.arccosh: 'arccosh',
    UnOp.arctgh: 'arctgh',
    VectorBinOp.Add: 'vAdd',
    VectorBinOp.Sub: 'vSub',
    VectorBinOp.MultByNum: 'vMulNum',
    VectorBinOp.DotProd: 'vDP',
    VectorBinOp.Mult: 'vMul',
    VectorBinOp.TensMult: 'vTMul',
    VectorUnOp.USub: 'vUSub',
    VectorUnOp.Transpose: 'vT',
    Compare.Eq: '=',
    Compare.NotEq: '!=',
    Compare.Lt: '<',
    Compare.LtE: '<=',
    Compare.Gt: '>',
    Compare.GtE: '>='
}


def raise_(ex):
    raise ex

oper_actions = {
    BinOp.Add: lambda x, y: x + y,
    BinOp.Sub: lambda x, y: x - y,
    BinOp.Mult: lambda x, y: x * y,
    BinOp.Div: lambda x, y: x / y if y != 0 else raise_(ZeroDivisionError),
    BinOp.Pow: lambda x, y: x ** y,
    UnOp.sin: lambda x: math.sin(x),
    UnOp.cos: lambda x: math.cos(x),
    UnOp.abs: lambda x: math.fabs(x),
    UnOp.tg: lambda x: math.tan(x),
    UnOp.ctg: lambda x: 1.0/math.tan(x),
    UnOp.exp: lambda x: math.exp(x),
    UnOp.sinh: lambda x: math.sinh(x),
    UnOp.cosh: lambda x: math.cosh(x),
    UnOp.tgh: lambda x: math.tanh(x),
    UnOp.ctgh: lambda x: 1.0/math.tanh(x),
    UnOp.arcsin: lambda x: math.asin(x),
    UnOp.arccos: lambda x: math.acos(x),
    UnOp.arctg: lambda x: math.atan(x),
    UnOp.arcctg: lambda x: math.pi/2-math.atan(x),
    UnOp.arcsinh: lambda x: math.asinh(x),
    UnOp.arccosh: lambda x: math.acosh(x),
    UnOp.arctgh: lambda x: math.atanh(x),
    UnOp.log2: lambda x: math.log2(x),
    UnOp.log10: lambda x: math.log10(x),
    UnOp.ln: lambda x: math.log(x),
    BinOp.log: lambda x, y: math.log(x, y),
    BinOp.mod: lambda x, y: x % y
}


def vector_bin_op_add(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and
            node2.value.is_operation() and node2.value.operation == VectorOp.Vector and
            node1.value.shape == node2.value.shape):
        raise ValueError('Vector addition error')

    new_node = Node(NodeValue(operation=VectorOp.Vector))
    for i in range(len(node1.children)):
        child_node1 = node1.children[i]
        child_node2 = node2.children[i]

        if child_node1.value.shape == ():
            new_node.add_child(Node(NodeValue(operand=Operand(numeric_value=child_node1.value.operand.numeric_value +
                                                                            child_node2.value.operand.numeric_value))))
        else:
            new_node.add_child(vector_bin_op_add(child_node1, child_node2))

    new_node.value.shape = node1.value.shape
    return new_node
oper_actions[VectorBinOp.Add] = vector_bin_op_add


def vector_binop_sub(node1, node2):
    new_node = vector_bin_op_add(node1, vector_un_op_u_sub(node2))
    new_node.value.shape = node1.value.shape
    return new_node
oper_actions[VectorBinOp.Sub] = vector_binop_sub


def vector_bin_op_mul_by_num(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and node2.value.shape == () or
            node2.value.is_operation() and node2.value.operation == VectorOp.Vector and node1.value.shape == ()):
        raise ValueError('Vector multiplication by number error')

    if node1.value.shape != ():
        node1, node2 = node2, node1

    new_node = Node(NodeValue(operation=VectorOp.Vector))
    for i in range(len(node2.children)):
        child_node = node2.children[i]

        if child_node.value.shape == ():
            new_node.add_child(Node(NodeValue(operand=Operand(numeric_value=node1.value.operand.numeric_value *
                                                                                  child_node.value.operand.numeric_value))))
        else:
            new_node.add_child(vector_bin_op_mul_by_num(node1, child_node))

    new_node.value.shape = node2.value.shape
    return new_node
oper_actions[VectorBinOp.MultByNum] = vector_bin_op_mul_by_num


def vector_bin_op_dot_product(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and node1.value.shape[0] == 1 and
            node2.value.is_operation() and node2.value.operation == VectorOp.Vector and node1.value.shape[0] == 1 and
            node1.value.shape == node2.value.shape):
        raise ValueError('Vector dot product error')

    if node1.children[0].value.is_operation():
        node1 = node1.children[0]
    if node2.children[0].value.is_operation():
        node2 = node2.children[0]

    val = 0
    for i in range(len(node1.children)):
        child_node1 = node1.children[i]
        child_node2 = node2.children[i]

        if child_node1.value.operand.is_numeric_value():
            val += child_node1.value.operand.numeric_value * child_node2.value.operand.numeric_value

    new_node = Node(NodeValue(operand=Operand(numeric_value=val)))
    return new_node
oper_actions[VectorBinOp.DotProd] = vector_bin_op_dot_product


def vector_bin_op_get_ind(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and node2.value.shape == ()):
        raise ValueError('Vector multiplication by number error')
    if node1.value.shape[0] == 1:
        if not (0 <= node2.value.operand.numeric_value - 1 < len(node1.children[0].children)):
            raise ValueError('Wrong index parameter')

        new_node = node1.children[0].children[node2.value.operand.numeric_value - 1]
        if new_node.value.shape != ():
            tmp_node = Node(NodeValue(operation=VectorOp.Vector))
            tmp_node.add_child(new_node)
            tmp_node.value.shape = new_node.value.shape

            new_node = tmp_node

    else:
        if not (0 <= node2.value.operand.numeric_value - 1 < len(node1.children)):
            raise ValueError('Wrong index parameter')

        new_node = node1.children[node2.value.operand.numeric_value - 1]
        if new_node.value.shape != ():
            tmp_node = Node(NodeValue(operation=VectorOp.Vector))
            tmp_node.add_child(new_node)
            tmp_node.value.shape = new_node.value.shape

            new_node = tmp_node

    if new_node.value.shape == (1, 1):
        new_node = new_node.children[0].children[0]
    return new_node

oper_actions[VectorBinOp.GetInd] = vector_bin_op_get_ind


def vector_un_op_u_sub(node):
    from sqc.src.Node import Node, NodeValue

    if not (node.value.is_operation() and node.value.operation == VectorOp.Vector):
        raise ValueError('Vector unary subtraction error')

    new_node = vector_bin_op_mul_by_num(Node(NodeValue(operand=Operand(numeric_value=-1))), node)

    new_node.value.shape = node.value.shape
    return new_node
oper_actions[VectorUnOp.USub] = vector_un_op_u_sub


def vector_un_op_transpose(node):
    from sqc.src.Node import Node, NodeValue

    if not (node.value.is_operation() and node.value.operation == VectorOp.Vector):
        raise ValueError('Vector transpose error')

    new_node = Node(NodeValue(operation=VectorOp.Vector))
    for i in range(node.value.shape[1]):
        new_node.add_child(Node(NodeValue(operation=VectorOp.Vector, shape=(1, node.value.shape[0]))))
    for i in range(node.value.shape[0]):
        for j in range(node.value.shape[1]):
            new_node.children[j].add_child(copy.copy(node.children[i].children[j]))

    new_node.value.shape = (node.value.shape[1], node.value.shape[0])
    return new_node
oper_actions[VectorUnOp.Transpose] = vector_un_op_transpose


def vector_bin_op_mul(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and
            node2.value.is_operation() and node2.value.operation == VectorOp.Vector and
            node1.value.shape[1] == node2.value.shape[0]):
        raise ValueError('Vector multiplication error')

    new_node = Node(NodeValue(operation=VectorOp.Vector))
    for i in range(node1.value.shape[0]):
        new_node.add_child(Node(NodeValue(operation=VectorOp.Vector, shape=(1, node2.value.shape[1]))))

    node2_t = vector_un_op_transpose(node2)

    for i in range(node1.value.shape[0]):
        for j in range(node2_t.value.shape[0]):
            new_node.children[i].add_child(vector_bin_op_dot_product(node1.children[i], node2_t.children[j]))

    new_node.value.shape = (node1.value.shape[0], node2.value.shape[1])

    if new_node.value.shape == (1, 1):
        new_node = new_node.children[0].children[0]
    return new_node
oper_actions[VectorBinOp.Mult] = vector_bin_op_mul


def vector_bin_op_tens_mul(node1, node2):
    from sqc.src.Node import Node, NodeValue

    if not (node1.value.is_operation() and node1.value.operation == VectorOp.Vector and
            node2.value.is_operation() and node2.value.operation == VectorOp.Vector):
        raise ValueError('Vector tensor multiplication error')

    new_node = Node(NodeValue(operation=VectorOp.Vector))
    for i in range(node1.value.shape[0]*node2.value.shape[0]):
        new_node.add_child(
            Node(NodeValue(operation=VectorOp.Vector, shape=(1, node1.value.shape[1]*node2.value.shape[1])))
        )

    for i1 in range(node1.value.shape[0]):
        for j1 in range(node1.value.shape[1]):
            for i2 in range(node2.value.shape[0]):
                for j2 in range(node2.value.shape[1]):
                    new_node.children[i1*node2.value.shape[0] + i2].add_child(Node(NodeValue(
                        operand=Operand(numeric_value=
                                        node1.children[i1].children[j1].value.operand.numeric_value*
                                        node2.children[i2].children[j2].value.operand.numeric_value
                                        )
                    )))

    new_node.value.shape = (node1.value.shape[0]*node2.value.shape[0],
                            node1.value.shape[1]*node2.value.shape[1])

    if new_node.value.shape == (1, 1):
        new_node = new_node.children[0].children[0]
    return new_node
oper_actions[VectorBinOp.TensMult] = vector_bin_op_tens_mul



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_empty(self):
        input_string = ''
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '')

    def test_binary_op_sum(self):
        input_string = '1+1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '2\n')

    def test_binary_op_sub(self):
        input_string = '1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '0\n')

    def test_binary_op_mul(self):
        input_string = '1*1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '1\n')

    def test_binary_op_div(self):
        input_string = '1/1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '1.0\n')

    def test_binary_op_pow(self):
        input_string = '1**1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '1\n')

    def test_two_ops(self):
        input_string = '1+1+1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '3\n')

    def test_two_ops2(self):
        input_string = '1+1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '1\n')

    def test_three_ops(self):
        input_string = '1/1*1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '0.0\n')

    def test_order_ops1(self):
        input_string = '1*2/8*4'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '1.0\n')

    def test_order_ops2(self):
        input_string = '1+2*3/4-5'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '-2.5\n')

    def test_order_pow(self):
        input_string = '1**2*3'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '3\n')

    def test_parentheses(self):
        input_string = '(1+1)*(2+2)/(3-4)+(5*6)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '22.0\n')

    def test_functions(self):
        input_string = 'mod(5,3)+log2(2)+log10(10)+log(3,3)+sin(arcsin(1))+cos(arccos(1))+tg(arctg(1))+ctg(arcctg(1))+exp(0)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, '10.0\n')

    def test_functions2(self):
        input_string = 'sinh(arcsinh(1))+cosh(arccosh(1))+tgh(arctgh(z))+ctgh(x)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, 'Add\n\tAdd\n\t\t2.0\n\t\ttgh\n\t\t\tarctgh\n\t\t\t\tz\n\tctgh\n\t\tx\n')

    def test_with_args(self):
        input_string = '2+3+x'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, 'Add\n\t5\n\tx\n')

    def test_with_args2(self):
        input_string = 'a*b/a'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = str(tree)
        self.assertEquals(output_string, 'Div\n\tMult\n\t\ta\n\t\tb\n\ta\n')


if __name__ == '__main__':
    unittest.main()



import ast
import copy
import os
from termcolor import colored

from sqc.src.Node import Node, NodeValue
from sqc.src.Operand import Operand
from sqc.src.misc import get_index

from sqc.src.operations import BinOp, UnOp, functions, Compare
from sqc.src.operations import ast_bin_ops, ast_un_ops, ast_compare_ops
from sqc.src.operations import oper_symbols
from sqc.src.operations import VectorUnOp, VectorBinOp, VectorOp
from sqc.src.Constant import constants


class Tree:
    """
    Abstract syntax tree class.
    """

    def __init__(self, var_shapes=None, ast_tree=None, expression=None, root=None):
        """
            Copies ast.

            :param ast_tree:    Built-in abstract syntax tree.
            :type ast_tree:     ast.Module
            :param expression:  Arithmetic expression
            :type expression:   string
            :return:            Rebuilt abstract syntax tree.
            :rtype:             Tree
            """

        def df_walk(ast_node, parent):
            """
            Depth-first walk on the tree.

            :param ast_node:            Built-in ast node.
            :type ast_node:             ast.Module, ast.Expr, ast.BinOp, UnaryOp, ast.Name, ast.Num, ast.Call, ast.List
            :return:
            :rtype:                     Node
            """

            new_node = None
            if isinstance(ast_node, ast.Module):
                if len(ast_node.body) > 0:
                    new_node = df_walk(ast_node.body[0], ast_node)

            elif isinstance(ast_node, ast.Expr):
                new_node = df_walk(ast_node.value, ast_node)

            elif isinstance(ast_node, ast.BinOp):
                new_node = Node(NodeValue(operation=ast_bin_ops[type(ast_node.op).__name__]))
                
                new_node.add_child(df_walk(ast_node.left, ast_node))
                new_node.add_child(df_walk(ast_node.right, ast_node))

            elif isinstance(ast_node, ast.UnaryOp):
                new_node = Node(NodeValue(operation=ast_un_ops[type(ast_node.op).__name__]))

                new_node.add_child(df_walk(ast_node.operand, ast_node))

            elif isinstance(ast_node, ast.Compare):
                new_node = Node(NodeValue(operation=ast_compare_ops[type(ast_node.ops[0]).__name__]))

                new_node.add_child(df_walk(ast_node.left, ast_node))
                new_node.add_child(df_walk(ast_node.comparators[0], ast_node))
            elif isinstance(ast_node, ast.Name):
                new_node = Node(NodeValue(operand=Operand(variable=ast_node.id)))

                if ast_node.id in constants.keys():
                    new_node.value.shape = ()
                    new_node.value.operand.shape = ()
                elif var_shapes is None:
                    new_node.value.shape = ()
                    new_node.value.operand.shape = ()
                elif ast_node.id in var_shapes:
                    new_node.value.shape = var_shapes[ast_node.id]
                    new_node.value.operand.shape = var_shapes[ast_node.id]
                else:
                    raise ValueError('Missing shape definition for variable ' + ast_node.id)

            elif isinstance(ast_node, ast.Num):
                new_node = Node(NodeValue(operand=Operand(numeric_value=ast_node.n)))

            elif isinstance(ast_node, ast.Call):
                if len(ast_node.args) == 2:
                    new_node = Node(NodeValue(operation=ast_bin_ops[ast_node.func.id]))

                    new_node.add_child(df_walk(ast_node.args[0], ast_node))
                    new_node.add_child(df_walk(ast_node.args[1], ast_node))
                elif len(ast_node.args) == 1:
                    new_node = Node(NodeValue(operation=ast_un_ops[ast_node.func.id]))
                    new_node.add_child(df_walk(ast_node.args[0], ast_node))
            elif isinstance(ast_node, ast.List):
                if len(ast_node.elts) == 0:
                    raise ValueError('Empty vector error')

                new_node = Node(NodeValue(operation=VectorOp.Vector))

                for node in ast_node.elts:
                    new_node.add_child(df_walk(node, ast_node))

                new_node.set_shape()
                if new_node.value.shape[0] == 1 and new_node.children[0].value.shape == () and not\
                   (parent is not None and isinstance(parent, ast.List)):
                    tmp_node = Node(NodeValue(operation=VectorOp.Vector))
                    tmp_node.add_child(new_node)
                    tmp_node.value.shape = new_node.value.shape

                    new_node = tmp_node

                if new_node.value.shape == (1, 1) and len(new_node.children[0].children) > 0:
                    new_node = new_node.children[0].children[0]

            if new_node is not None:
                new_node.set_shape()
            return new_node
        if root is not None:
            self.root = root
        else:
            if ast_tree is None:
                try:
                    ast_tree = ast.parse(expression)
                except SyntaxError:
                    print('Syntax error. Unexpected symbols in AST')
                    self.root = None
                else:
                    self.root = df_walk(ast_tree, None)
                    if self.root is None:
                        self.root = Node(None)

        self.action = Compare.Eq
        self.current_node = self.root

    def __str__(self):
        def dfs(node, depth):
            result = ''
            if node.value.is_operation():
                result += '\t' * depth + str(node.value) + '\n'
                for child in node.children:
                    result += dfs(child, depth + 1)
            else:
                result = '\t' * depth + str(node.value) + '\n'

            return result

        output = ''
        if self.root.value is not None:
            output = dfs(self.root, 0)

        return output

    def __eq__(self, other):
        return self.root.eq_subtrees(other.root)

    def node_list(self):
        def dfs(node):
            nodes = [node]
            for child in node.children:
                nodes.extend(dfs(child))
            return nodes

        return dfs(self.root)

    def tree_calculate(self):
        def dfs(node):
            for child in node.children:
                dfs(child)
            node.calculate()

        if self.root.value is not None:
            dfs(self.root)

    def to_expression(self, color=None):
        """
        Transform tree to expression.
        :return:    Expression string.
        :rtype:     string
        """

        def dfs(node):
            if node.value.is_operation():
                if isinstance(node.value.operation, VectorOp):
                    result = ''
                    if node.value.shape[0] != 1 or node.parent is not None and\
                       node.parent.value.is_operation() and node.parent.value.operation == VectorOp.Vector:
                        result += '['

                    for i in range(len(node.children)):
                        child_node = node.children[i]

                        if i != 0:
                            result += ', '

                        result += dfs(child_node)

                    if node.value.shape[0] != 1 or node.parent is not None and \
                       node.parent.value.is_operation() and node.parent.value.operation == VectorOp.Vector:
                        result += ']'
                else:
                    if isinstance(node.value.operation, UnOp):
                        result = oper_symbols[node.value.operation]
                    else:
                        result = ''
                    if node.value.operation in functions:
                        result = oper_symbols[node.value.operation] + '('

                    for i in range(len(node.children)):
                        child = node.children[i]
                        if child.value.is_operation() and child.value.operation not in functions:
                            result += '(' + dfs(child) + ')'
                            if node.value.is_operation() and node.value.operation in functions:
                                if i != len(node.children) - 1:
                                    result += ', '
                        else:
                            result += dfs(child)
                            if node.value.operation in functions:
                                if i != len(node.children) - 1:
                                    result += ', '

                        if i != len(node.children) - 1 and node.value.operation not in functions:
                            result += oper_symbols[node.value.operation]

                    if node.value.operation in functions:
                        result += ')'
            else:
                if node.value.operand.is_numeric_value() and node.value.operand.numeric_value < 0 and\
                   node.parent is not None and\
                   node.parent.value.operation != VectorOp.Vector:
                    result = '(' + str(node.value) + ')'
                else:
                    result = str(node.value)

            if color is not None and node.subroot:
                result = colored(result, color)

            return result

        output = ''
        if self.root.value is not None:
            output = dfs(self.root)

        return output

    def next_node(self):
        if len(self.current_node.children) == 0:
            parent = self.current_node.parent
            while parent is not None and\
                    get_index(parent.children, self.current_node) == len(parent.children) - 1:
                self.current_node = parent
                parent = parent.parent

            if parent is None:
                self.current_node = self.root
                return None
            else:
                self.current_node = parent.children[parent.children.index(self.current_node) + 1]
        else:
            self.current_node = self.current_node.children[0]

        return self.current_node

    def get_copy(self):
        """
        Return copy of the tree.
        :return:
        :rtype:     Tree
        """
        def dfs(orig_node, copy_parent):
            new_node = Node(copy.copy(orig_node.value), orig_node.subroot)
            copy_parent.add_child(new_node)
            for child in orig_node.children:
                dfs(child, new_node)

        tmp_root = Node(None)
        dfs(self.root, tmp_root)
        if len(tmp_root.children) > 0:
            tmp_root = tmp_root.children[0]

        return Tree(var_shapes=None, root=tmp_root)


class Rules:
    def __init__(self):
        self.data = []
        with open(os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/rules.txt', 'r') as f:
            rules_lines = f.read().split("\n")
            for id,line in enumerate(rules_lines):
                self.data.append((id, self.get_rule_from_string(line)))

    def __iter__(self):
        for it in self.data:
            yield it

    @staticmethod
    def get_rule_from_string(string):
        left = None
        right = None
        action = None
        if ':' in string:
            left, right = string.split(':')
            action = Compare.Eq
        elif '<=' in string:
            left, right = string.split('<=')
            action = Compare.LtE
        elif '<' in string:
            left, right = string.split('<')
            action = Compare.Lt
        elif '>=' in string:
            left, right = string.split('>=')
            action = Compare.GtE
        elif '>' in string:
            left, right = string.split('>')
            action = Compare.Gt

        return Tree(expression=left), Tree(expression=right), action

    @staticmethod
    def apply_rule(node, rule):
        unify_result = node.unify(rule[0].root)
        if unify_result is not None:
            node.replace(rule)

    def get_possible_rules(self, node, tree):
        possible_actions = None
        result = []
        node_copy = copy.copy(node)
        if tree.action == Compare.Eq:
            possible_actions = [Compare.Eq, Compare.LtE, Compare.Lt, Compare.GtE, Compare.Gt]
        if tree.action == Compare.Gt or tree.action == Compare.GtE:
            possible_actions = [Compare.Eq, Compare.GtE, Compare.Gt]
        if tree.action == Compare.Lt or tree.action == Compare.LtE:
            possible_actions = [Compare.Eq, Compare.LtE, Compare.Lt]

        possible_rules = [(id,rule) for id, rule in self.data if rule[2] in possible_actions]
        for id, rule in possible_rules:
            if node.unify(rule[0].root) is not None:
                node.replace(rule)
                node.subroot = True

                copy_tree = tree.get_copy()
                copy_tree.action = tree.action
                result.append([copy_tree, rule, id])

                node.value = node_copy.value
                node.subroot = False
                node.children = node_copy.children

        for child in node.children:
            result.extend(self.get_possible_rules(child, tree))

        return result



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_empty_expression(self):
        input_string = ''
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, '')

    def test_unary_op(self):
        input_string = '+1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'UAdd\n\t1\n')

    def test_binary_op_sum(self):
        input_string = '1+1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Add\n\t1\n\t1\n')

    def test_binary_op_sub(self):
        input_string = '1-1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Sub\n\t1\n\t1\n')

    def test_binary_op_mul(self):
        input_string = '1*1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Mult\n\t1\n\t1\n')

    def test_binary_op_div(self):
        input_string = '1/1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Div\n\t1\n\t1\n')

    def test_binary_op_pow(self):
        input_string = '1**1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Pow\n\t1\n\t1\n')

    def test_two_ops(self):
        input_string = '1+1+1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Add\n\tAdd\n\t\t1\n\t\t1\n\t1\n')

    def test_two_ops2(self):
        input_string = '1+1-1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Sub\n\tAdd\n\t\t1\n\t\t1\n\t1\n')

    def test_three_ops(self):
        input_string = '1/1*1-1'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Sub\n\tMult\n\t\tDiv\n\t\t\t1\n\t\t\t1\n\t\t1\n\t1\n')

    def test_order_ops1(self):
        input_string = '1*2/3*4'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Mult\n\tDiv\n\t\tMult\n\t\t\t1\n\t\t\t2\n\t\t3\n\t4\n')

    def test_order_ops2(self):
        input_string = '1+2*3/4-5'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Sub\n\tAdd\n\t\t1\n\t\tDiv\n\t\t\tMult\n\t\t\t\t2\n\t\t\t\t3\n\t\t\t4\n\t5\n')

    def test_order_pow(self):
        input_string = '1**2*3'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Mult\n\tPow\n\t\t1\n\t\t2\n\t3\n')

    def test_parentheses(self):
        input_string = '(1+1)*(2+2)/(3-4)+(5*6)'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Add\n\tDiv\n\t\tMult\n\t\t\tAdd\n\t\t\t\t1\n\t\t\t\t1\n\t\t\tAdd\n\t\t\t\t2\n\t\t\t\t2\n\t\tSub\n\t\t\t3\n\t\t\t4\n\tMult\n\t\t5\n\t\t6\n')

    # need to check this
    def test_parentheses2(self):
        input_string = '((a+1)*c+b*a+b)/(b+c)'
        output_string = str(Tree(expression=input_string))
        self.assertEquals(output_string, 'Div\n\tAdd\n\t\tAdd\n\t\t\tMult\n\t\t\t\tAdd\n\t\t\t\t\ta\n\t\t\t\t\t1\n\t\t\t\tc\n\t\t\tMult\n\t\t\t\tb\n\t\t\t\ta\n\t\tb\n\tAdd\n\t\tb\n\t\tc\n')

if __name__ == '__main__':
    unittest.main()



from enum import Enum
import math


class Constant(Enum):
    i = 1
    pi = 2
    e = 3

constants = {
    'i': Constant.i,
    'pi': Constant.pi,
    'e': Constant.e
}

constant_values = {
    Constant.pi: math.pi,
    Constant.e: math.e
}

constant_symbols = {
    Constant.pi: 'pi',
    Constant.e: 'e'
}



import unittest

from sqc.src.Tree import Tree, Rules
from sqc.src.operations import Compare


class Test(unittest.TestCase):
    def test_const1(self):
        input_string = '2*pi*i'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('i*i:-1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '(2*pi)*i')

    def test_const2(self):
        input_string = 'i*i'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('i*i:-1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '-1')

    def test_const3(self):
        input_string = 'ln(e)'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('ln(e):1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_const4(self):
        input_string = 'e**(pi*i)'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('e**(pi*i):-1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '-1')

    def test_mult_by_one(self):
        input_string = 'a*b'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('a:1*a')
        Rules.apply_rule(tree.root.children[0], rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '(1*a)*b')

    def test_trig(self):
        input_string = 'sin(a)**2+sin(a)**2'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('sin(x)**2+sin(x)**2:1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_1(self):
        input_string = '(a+1)**2'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('(a+b)**2:a**2+2*a*b+b**2')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, Tree(expression='a**2+2*a*1+1**2').to_expression())

    def test_Lt_apply(self):
        input_string = '1+x'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('1+x<e**x')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, Tree(expression='e**x').to_expression())

    def test_LtE_apply(self):
        input_string = 'sin(t)'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('sin(x)<=1')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, Tree(expression='1').to_expression())

    def test_Gt_apply(self):
        input_string = 'e**z'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('e**x>1+x')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, Tree(expression='1+z').to_expression())

    def test_GtE_apply(self):
        input_string = 'x**2'
        tree = Tree(expression=input_string)

        rule = Rules.get_rule_from_string('x**2>=0')
        Rules.apply_rule(tree.root, rule)

        output_string = tree.to_expression()
        self.assertEquals(output_string, Tree(expression='0').to_expression())

if __name__ == '__main__':
    unittest.main()



import unittest


def run_tests():
    loader = unittest.TestLoader()
    suite = loader.discover(start_dir='sqc.test_package.tests', pattern='test_*.py')

    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)


import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_empty(self):
        with self.assertRaises(Exception):
            Tree(expression='[]')

    def test_uni_vector(self):
        input_string = '[1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_empty_double_vector(self):
        input_string = '[[2], [1]]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2], [1]]')

    def test_vector1(self):
        input_string = '[x]'

        var_shape = {'x': ()}

        tree = Tree(expression=input_string, var_shapes=var_shape)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'x')

    def test_vector2(self):
        input_string = '[x, y]'

        var_shape = {'x': (), 'y': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[x, y]')

    def test_vector3(self):
        input_string = '[[1, x], [2, y]]'

        var_shape = {'x': (), 'y': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[1, x], [2, y]]')

    def test_vector_ops_Add_1(self):
        input_string = 'vAdd([3], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Add_2(self):
        input_string = 'vAdd([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[4, 10]')

    def test_vector_ops_Add_3(self):
        input_string = 'vAdd([3], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Add_4(self):
        input_string = 'vAdd([3, 2], [1, 5])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[4, 7]')

    def test_vector_ops_Add_5(self):
        input_string = 'vAdd([[3, 5], [1, 6], [3, 7]], [[3, 5], [1, 6], [3, 7]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[6, 10], [2, 12], [6, 14]]')

    def test_vector_ops_Sub_1(self):
        input_string = 'vSub([3], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Sub_2(self):
        input_string = 'vSub([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2, -2]')

    def test_vector_ops_Sub_3(self):
        input_string = 'vSub([[3]], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Sub_4(self):
        input_string = 'vSub([3, 2], [1, 5])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2, -3]')

    def test_vector_ops_Sub_5(self):
        input_string = 'vSub([[3, 5], [1, 6], [3, 7]], [[3, 5], [1, 6], [3, 7]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[0, 0], [0, 0], [0, 0]]')

    def test_vector_ops_MulNum_1(self):
        input_string = 'vMulNum(3, [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_MulNum_2(self):
        input_string = 'vMulNum([3, 4], -2)'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[-6, -8]')

    def test_vector_ops_MulNum_3(self):
        input_string = 'vMulNum(1, [[3], [1]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[3], [1]]')

    def test_vector_ops_MulNum_4(self):
        input_string = 'vMulNum(-2, [[3, 2], [1, 5]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[-6, -4], [-2, -10]]')

    def test_vector_ops_MulNum_5(self):
        input_string = 'vMulNum([1], [[3, 2], [1, 5]])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[3, 2], [1, 5]]')

    def test_vector_ops_DP_1(self):
        input_string = 'vDP([3], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_DP_2(self):
        input_string = 'vDP([3, 4], [1, 6])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '27')

    def test_vector_ops_DP_3(self):
        input_string = 'vDP([3, 4], [1, 6])-27'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0')

    def test_vector_ops_DP_4(self):
        input_string = 'vDP([3, 4], [1, 6])/10'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2.7')

    def test_vector_ops_DP_5(self):
        input_string = 'vDP([1], [[3, 2], [1, 5]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_vector_ops_DP_6(self):
        input_string = 'vDP([1], [3, 2])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_vector_ops_Get_1(self):
        input_string = 'vGet([3], 1)'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Get_2(self):
        input_string = 'vGet([3, 4], 2)'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '4')

    def test_vector_ops_Get_3(self):
        input_string = 'vGet([[3], [4]], 1)'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_vector_ops_Get_4(self):
        input_string = 'vGet([[3, 4], [1, 6]], 2)'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[1, 6]')

    def test_vector_ops_Get_5(self):
        input_string = 'vGet(1, [[3, 2], [1, 5]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_vector_ops_Get_6(self):
        input_string = 'vGet([3, 2], 1.0)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_vector_ops_Get_7(self):
        input_string = 'vGet([3, 2], 3)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_vector_ops_USub_1(self):
        input_string = 'vUSub([[3], [4]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[-3], [-4]]')

    def test_vector_ops_USub_2(self):
        input_string = 'vUSub([1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_USub_3(self):
        input_string = 'vUSub(1)'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_nested_ops1(self):
        input_string = 'vAdd(vAdd([1, 2], [3, 4]), vAdd([5, 6], [7, 8]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[16, 20]')

    def test_nested_ops2(self):
        input_string = 'vAdd(vAdd([1, 2], [3, 4]), vSub([5, 6], [7, 8]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2, 4]')

    def test_nested_ops3(self):
        input_string = 'vAdd(vMulNum(2, [3, 4]), vSub([5, 6], [7, 8]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[4, 6]')

    def test_nested_ops4(self):
        input_string = 'vSub(vMulNum(2, [3, 4]), vSub([5, 6], [7, 8]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[8, 10]')

    def test_nested_ops5(self):
        input_string = 'vAdd(vGet([[1, 2], [3, 4]], 1), [1, 2])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[2, 4]')

    def test_nested_ops6(self):
        input_string = 'vGet([1, 2], 2)-2'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0')

    def test_nested_ops7(self):
        input_string = 'vDP(vGet([[1, 2], [1, 2]], 2), vAdd([1, 2], [3, 4]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '16')

    def test_nested_ops8(self):
        input_string = 'vUSub(vMulNum(-1, [1, 2]))'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[1, 2]')

    def test_nested_ops9(self):
        input_string = 'vAdd(vGet([[3], [1]], 1), [-3])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_nested_ops10(self):
        input_string = 'vAdd([1, 2], vAdd([1, 2, 3], [4, 5, 6]))'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_nested_ops11(self):
        input_string = 'vAdd([1], [[1]])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_nested_ops12(self):
        input_string = 'vAdd(vDP([3], [4]), [5])'
        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)
            
            tree.tree_calculate()

    def test_nested_ops13(self):
        input_string = 'vMulNum([3], [1, 2])'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[3, 6]')

    def test_transpose1(self):
        input_string = 'vT([1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_transpose2(self):
        input_string = 'vT([1, 2])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[1], [2]]')

    def test_transpose3(self):
        input_string = 'vT([[1], [2]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[1, 2]')

    def test_transpose4(self):
        input_string = 'vT([[1, 2], [3, 4]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[1, 3], [2, 4]]')

    def test_transpose5(self):
        input_string = 'vAdd(vT([[1, 2, 3], [3, 4, 5]]), [[1, 3], [2, 4], [3, 5]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2, 6], [4, 8], [6, 10]]')

    def test_mult1(self):
        input_string = 'vMul([1], [[2]])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_mult2(self):
        input_string = 'vMul([1], [2])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_mult3(self):
        input_string = 'vMul([1, 2], [[2], [0]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2')

    def test_mult4(self):
        input_string = 'vMul([[1], [2]], [[2]])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_mult5(self):
        input_string = 'vMul([[1, 2], [3, 4]], [[2], [2]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[6], [14]]')

    def test_mult6(self):
        input_string = 'vMul([[1, 2], [3, 4]], [[2, 1], [2, 0]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[6, 1], [14, 3]]')

    def test_mult7(self):
        input_string = 'vMul([[1, 2], [3, 4]], [[2, 1, 3], [2, 4, 5]])'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[6, 9, 13], [14, 19, 29]]')

    def test_tens_mult1(self):
        input_string = 'vTMul([1], [[2]])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_tens_mult2(self):
        input_string = 'vTMul([1], [2])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_tens_mult3(self):
        input_string = 'vTMul([1, 2], [[2], [0]])'
        tree = Tree(expression=input_string)

        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2, 4], [0, 0]]')

    def test_tens_mult4(self):
        input_string = 'vTMul([[1], [2]], [[2]])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_tens_mult5(self):
        input_string = 'vTMul([[1, 2], [3, 4]], [[2], [2]])'
        tree = Tree(expression=input_string)

        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2, 4], [2, 4], [6, 8], [6, 8]]')

    def test_tens_mult6(self):
        input_string = 'vTMul([[1, 2], [3, 4]], [[2, 1], [2, 0]])'
        tree = Tree(expression=input_string)

        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[2, 1, 4, 2], [2, 0, 4, 0], [6, 3, 8, 4], [6, 0, 8, 0]]')

    def test_tens_mult7(self):
        input_string = 'vTMul([1, 2], [3, 4])'
        tree = Tree(expression=input_string)

        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[3, 4, 6, 8]')

    def test_tens_mult8(self):
        input_string = 'vTMul([[1], [2]], [[3], [4]])'
        tree = Tree(expression=input_string)

        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '[[3], [4], [6], [8]]')

if __name__ == '__main__':
    unittest.main()



import argparse

from termcolor import colored

from sqc.src.Tree import Tree
from sqc.test_package import test

from sqc.src.interface import Interface
from sqc.src.operations import oper_symbols


def run(st, is_advanced=False):
    log = open('log.txt', 'w')

    variable_shapes = Interface.get_variable_shapes()
    tree = Tree(var_shapes=variable_shapes, expression=st)
    expression = Interface(tree=tree)

    step = 0
    while tree is not None:
        if tree.root is None:
            break

        step += 1
        print()
        print('STEP ' + str(step))
        log.write('\n STEP ' + str(step) + ' ')

        if st.find('/0') >= 0 or st.find('%0') >= 0:
            print(colored('Zero division. Try again:', 'red'))
            answer = input()
            if answer == '0' or answer == 'exit':
                tree = None
            else:
                variable_shapes = Interface.get_variable_shapes()
                tree = Tree(expression=answer, var_shapes=variable_shapes)

            expression = Interface(tree=tree)
            step = 0
            continue

        try:
            tree.tree_calculate()
        except ZeroDivisionError:
            print(colored('Zero division. Try again:', 'red'))
            answer = input()
            if answer == '0' or answer == 'exit':
                tree = None
            else:
                variable_shapes = Interface.get_variable_shapes()
                tree = Tree(expression=answer, var_shapes=variable_shapes)

            expression = Interface(tree=tree)
            step = 0
            continue

        print('Tree:')
        print(tree)

        # log.write('Tree:')
        # log.write(tree)

        if not is_advanced:
            print(expression.show_last_k_trees())
            # print('Expression before applying rule on this step:' + oper_symbols[tree.action], tree.to_expression())
        else:
            print('>', tree.to_expression())

        log.write(tree.to_expression())

        exists_rule = expression.write_possible_rules(tree, is_advanced)
        if exists_rule is None:
            print('Try again:')
        elif not exists_rule:
            print('No possible rules. You can write new expression:')

        tree, new = expression.apply_rule(log, is_advanced)
        if new:
            expression = Interface(tree=tree)
            st = tree.to_expression()
            variable_shapes = expression.get_variable_shapes()
            tree = Tree(expression=st, var_shapes=variable_shapes)
            step = 0


def main():
    parser = argparse.ArgumentParser(description='Symbolic quantum expression shortener')
    parser.add_argument('-t', action='store_true', help='do unit test')
    parser.add_argument('--exp', help='Run program. Argument must be expression without spaces, with quotes')
    parser.add_argument('--f', help='Read expressions from your file.')
    parser.add_argument('-m', action='store_true', help='Advanced mode. Only if you know what is it.')

    args = parser.parse_args()

    if args.t:
        test.run_tests()
    elif args.exp is not None:
        run(args.exp, args.m)
    elif args.f is not None:
        expression = Interface()
        expression.read_from_file(args.f, args.m)
        if expression is not None:
            for exp in expression.data:
                run(exp, args.m)
            else:
                print(colored('File is empty', 'red'))

            expression.data.close()
    else:
        expression = input()
        run(st=expression, is_advanced=args.m)

if __name__ == '__main__':
    main()





import unittest

from sqc.src.Tree import Tree, Rules


class Test(unittest.TestCase):
    def test_empty_expression(self):
        input_string = ''
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '0')

    def test1(self):
        input_string = '(a+a)/2'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a/2)+(a/2)')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a+a)/2')

        tree, rule, id = result[2]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(2*a)/2')

    def test2(self):
        input_string = 'a*1/a'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'a*(1/a)')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'a/a')

        tree, rule, id = result[2]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(1*a)/a')

    def test3(self):
        input_string = 'a+b-b'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a+b)+(-b)')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'a+(b-b)')

        tree, rule, id = result[2]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b+a)-b')

    def test4(self):
        input_string = 'a*b/a'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '2')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'a*(b/a)')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b*a)/a')

    def test5(self):
        input_string = 'b/a*a/b'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '2')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b/a)*(a/b)')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(b/a))/b')

    def test6(self):
        input_string = '(a*(b+c))/(c+2*b-b)'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '8')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, 'a*((b+c)/((c+(2*b))-b))')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((a*b)+(a*c))/((c+(2*b))-b)')

        tree, rule, id = result[2]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((b+c)*a)/((c+(2*b))-b)')

        tree, rule, id = result[3]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(c+b))/((c+(2*b))-b)')

        tree, rule, id = result[4]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(b+c))/((c+(2*b))+(-b))')

        tree, rule, id = result[5]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(b+c))/(c+((2*b)-b))')

        tree, rule, id = result[6]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(b+c))/(((2*b)+c)-b)')

        tree, rule, id = result[7]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(a*(b+c))/((c+(b*2))-b)')

    def test7(self):
        input_string = '(b*b+b)/(b+1)'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '7')

        tree, rule, id = result[0]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((b*b)/(b+1))+(b/(b+1))')

        tree, rule, id = result[1]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b+(b*b))/(b+1)')

        tree, rule, id = result[2]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b*(b+1))/(b+1)')

        tree, rule, id = result[3]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(b*(b+1))/(b+1)')

        tree, rule, id = result[4]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((b*b)+b)/(b+1)')

        tree, rule, id = result[5]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((b**2)+b)/(b+1)')

        tree, rule, id = result[6]
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((b*b)+b)/(1+b)')



import os
import sys
import ast
from termcolor import colored

from sqc.src.Tree import Tree, Rules
from sqc.src.Constant import constants
from sqc.src.operations import Compare, oper_symbols


class Interface:
    def __init__(self, tree=None):
        self.trees_on_steps = []
        if tree is not None:
            self.trees_on_steps.append(tree)
            self.step = 0
        else:
            self.step = -1
        self.data = {}
        self.max_rules = 0

    def read_from_file(self, file_path, is_advanced=False):
        while not os.path.exists(file_path):
            if file_path == "exit":
                sys.exit()
            elif file_path == '0':
                return None

            print("Can not open file. Try again")
            if not is_advanced:
                print("Enter 'exit' command to exit")

            file_path = input()

        file = open(file_path, 'r')
        self.data = file

    @staticmethod
    def get_variable_shapes():
        print('Enter names of variables and their shapes in form')
        print('<name> <shape_tuple>')
        print('End input with empty line')

        result = {}
        line = input()
        while len(line) > 0:
            if line.find(' '):
                var, shape = line.split(' ')[0], ast.literal_eval(line.split(' ')[1])

                if var in result.keys():
                    print('Variable already defined. Try again.')
                elif not isinstance(shape, tuple):
                    print('Wrong shape type. Expected tuple. Try again.')
                elif var in constants.keys():
                    print('Variable name reserved for constant. Try again.')
                else:
                    result[var] = shape
            else:
                print('Space separated line expected. Try again.')

            line = input()

        return result

    def write_possible_rules(self, tree, is_advanced):
        if not is_advanced:
            print("Choose the rule or write new expression:")

        if self.step > 0:
            print('-1. Back')

        print('0. Exit')
        print()

        rules = Rules()
        possible_rules = rules.get_possible_rules(tree.root, tree)
        num = 1
        error = False
        possible_rules_ids = set([el[2] for el in possible_rules])
        for id in possible_rules_ids:
            current_rules = [(el[0], el[1]) for el in possible_rules if el[2] == id]
            rule = current_rules[0][1]
            if not is_advanced:
                print('Rule:', rule[0].to_expression(), oper_symbols[rule[2]], rule[1].to_expression())
            for new_tree, rule in current_rules:
                self.data[num] = (new_tree, rule)
                new_exp = new_tree.to_expression('yellow')

                if not is_advanced:
                    print(str(num) + '. Expression after applying this rule:', new_exp)
                else:
                    print(str(num) + '. >', rule[0].to_expression(), oper_symbols[rule[2]], rule[1].to_expression(), '|',
                          new_exp)

                new_exp = new_tree.to_expression()
                if new_exp.find('/0') >= 0 or new_exp.find('%0') >= 0:
                    print(colored('Zero division.', 'red'))
                    error = True
                num += 1

        self.max_rules = num - 1
        if error:
            return None
        else:
            return num > 1

    def apply_rule(self, log, is_advanced):
        ok = False
        while not ok:
            answer = input()
            if answer is not None and len(answer) > 0:
                try:
                    answer = int(answer)
                except ValueError:
                    try:  # this try is useless now
                        tree = Tree(expression=answer)
                    except ValueError:
                        raise ValueError('The expression includes unexpected symbols')
                    else:
                        if tree.root is None:
                            print('Try again:')
                        else:
                            log.write('\n\n New expression: ' + answer)
                            return tree, True
                else:
                    if answer == 0:
                        print(colored('Step-by-step solution successfully saved in file log.txt', 'green'))
                        log.close()
                        return None, False
                    if answer == -1 and self.step > 0:
                        self.trees_on_steps.remove(self.trees_on_steps[self.step])
                        self.step -= 1
                        return self.trees_on_steps[self.step], False
                    elif 0 < answer <= self.max_rules:
                        if not is_advanced:
                            print('Leave your note for this step: ')
                        else:
                            print('note: ')

                        note = input()
                        if note is not None and len(note) > 0:
                            log.write('  # ' + note)

                        current_tree, current_rule = self.data[answer]
                        if current_tree.action != Compare.Gt and current_tree.action != Compare.Lt:
                            current_tree.action = current_rule[2]
                        self.trees_on_steps.append(current_tree)
                        self.step += 1
                        return current_tree, False
                    else:
                        print(colored('There are only', self.max_rules, 'rules. Try again.', 'red'))

    def show_last_k_trees(self, k=2):
        expr = 'Expression before applying rule on this step:'
        if self.step < k:
            k = self.step+1
        for i in range(k):
            tree = self.trees_on_steps[self.step+1-k+i]
            expr += oper_symbols[tree.action] + tree.to_expression()
        return expr



import unittest

from sqc.src.Tree import Tree, Rules


class Test(unittest.TestCase):
    def test_empty_expression(self):
        input_string = ''
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '0')

    def test1(self):
        input_string = '(a+a)/2'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a+c)/b=(a/b)+(c/b)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')

        tree, rule, id = result[2]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+a=2*a')

    def test2(self):
        input_string = 'a*1/a'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)/c=a*(b/c)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*1=a')

        tree, rule, id = result[2]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

    def test3(self):
        input_string = 'a+b-b'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '3')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a-b=a+(-b)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a+b)-c=a+(b-c)')

        tree, rule, id = result[2]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')

    def test4(self):
        input_string = 'a*b/a'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '2')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)/c=a*(b/c)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

    def test5(self):
        input_string = 'b/a*a/b'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '2')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)/c=a*(b/c)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

    def test6(self):
        input_string = '(a*(b+c))/(c+2*b-b)'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '8')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)/c=a*(b/c)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*(b+c)=(a*b)+(a*c)')

        tree, rule, id = result[2]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

        tree, rule, id = result[3]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')

        tree, rule, id = result[4]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a-b=a+(-b)')

        tree, rule, id = result[5]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a+b)-c=a+(b-c)')

        tree, rule, id = result[6]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')

        tree, rule, id = result[7]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

    def test7(self):
        input_string = '(b*b+b)/(b+1)'
        tree = Tree(expression=input_string)
        rules = Rules()
        result = rules.get_possible_rules(tree.root, tree)
        output_string = str(len(result))
        self.assertEquals(output_string, '7')

        tree, rule, id = result[0]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a+c)/b=(a/b)+(c/b)')

        tree, rule, id = result[1]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')

        tree, rule, id = result[2]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)+a=a*(b+1)')

        tree, rule, id = result[3]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, '(a*b)+b=b*(a+1)')

        tree, rule, id = result[4]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*b=b*a')

        tree, rule, id = result[5]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a*a=a**2')

        tree, rule, id = result[6]
        output_string = rule[0].to_expression() + '=' + rule[1].to_expression()
        self.assertEquals(output_string, 'a+b=b+a')



import copy

from sqc.src.Operand import Operand

from sqc.src.operations import BinOp, UnOp, VectorUnOp, VectorBinOp, VectorOp
from sqc.src.operations import oper_actions


class Node:
    """
    Node class.
        value:      NodeValue object
        children:   list of outgoing nodes
    """

    def __init__(self, value, subroot=False):
        self.value = value
        self.children = []
        self.parent = None
        self.subroot = subroot

    def __eq__(self, other):
        return self.value == other.value and len(self.children) == len(other.children)

    def __str__(self):
        return str(self.value)

    def __copy__(self):
        result = Node(copy.copy(self.value))
        for child in self.children:
            result.add_child(copy.copy(child))
        return result

    def add_child(self, node):
        node.parent = self
        self.children.append(node)

    def eq_subtrees(self, other):
        """
        Recursive function for checking whether subtrees are equal.

        :param other:   another node
        :type other:    Node
        :return:        True if corresponding subtrees are equal, False o/w.
        :rtype:         bool
        """

        result = self == other
        if result:
            for i in range(len(self.children)):
                child = self.children[i]
                other_child = other.children[i]

                result &= child.eq_subtrees(other_child)
                if not result:
                    break
        return result

    def calculate(self):
        """
        Tries to calculate children. (for binary trees only)
        :return:
        :rtype:
        """
        def all_children_are_numeric_values(node):
            result = True
            if node.value.is_operand() and not node.value.operand.is_numeric_value():
                result = False

            if result:
                for child_node in node.children:
                    result &= all_children_are_numeric_values(child_node)

                    if not result:
                        break

            return result

        if self.value.is_operation():
            if isinstance(self.value.operation, BinOp) and all_children_are_numeric_values(self):
                values = [child.value.operand.numeric_value for child in self.children]
                value = oper_actions[self.value.operation](values[0], values[1])

                self.children = []
                self.value = NodeValue(operand=Operand(numeric_value=value))
                self.value.shape = ()
                if isinstance(self.value.operand.numeric_value, complex):
                    complex_number = self.value.operand.numeric_value

                    self.value = NodeValue(operation=BinOp.Add)

                    self.add_child(Node(NodeValue(operand=Operand(numeric_value=complex_number.real))))
                    self.children[0].value.shape = ()

                    self.add_child(Node(NodeValue(operation=BinOp.Mult)))
                    self.children[1].value.shape = ()

                    self.children[1].add_child(Node(NodeValue(operand=Operand(numeric_value=complex_number.imag))))
                    self.children[1].children[0].value.shape = ()

                    self.children[1].add_child(Node(NodeValue(operand=Operand(variable='i'))))
                    self.children[1].children[1].value.shape = ()

            elif isinstance(self.value.operation, UnOp) and all_children_are_numeric_values(self):
                values = [child.value.operand.numeric_value for child in self.children]
                if self.value.operation == UnOp.UAdd:
                    self.children = []
                    self.value = NodeValue(operand=Operand(numeric_value=values[0]))
                elif self.value.operation == UnOp.USub:
                    self.children = []
                    self.value = NodeValue(operand=Operand(numeric_value=-values[0]))
                else:
                    value = oper_actions[self.value.operation](values[0])

                    self.children = []
                    self.value = NodeValue(operand=Operand(numeric_value=value))
                self.value.shape = ()

            elif isinstance(self.value.operation, VectorBinOp):
                if all_children_are_numeric_values(self):
                    new_node = oper_actions[self.value.operation](self.children[0], self.children[1])
                    new_node.value.shape = self.value.shape
                    self.value = new_node.value
                    self.children = new_node.children

            elif isinstance(self.value.operation, VectorUnOp):
                if all_children_are_numeric_values(self):
                    new_node = oper_actions[self.value.operation](self.children[0])
                    new_node.value.shape = self.value.shape
                    self.value = new_node.value
                    self.children = new_node.children

    def unify(self, rule_node):
        if self.value is None:
            return None

        if self.value.is_operand() and not rule_node.value.is_operand() or\
           (self.value.is_operation() and self.value.operation == VectorOp):
            return None

        result_data = {}
        if self.value.is_operation():
            if rule_node.value.is_operation():
                if len(self.children) != len(rule_node.children) or\
                   self.value.operation != rule_node.value.operation:
                    return None
                for i in range(len(self.children)):
                    inter_result = self.children[i].unify(rule_node.children[i])

                    if inter_result is None:
                        return None

                    for key, value in inter_result.items():
                        if key in result_data:
                            if not value.eq_subtrees(result_data[key]):
                                return None
                        else:
                            result_data[key] = value
            else:
                if not rule_node.value.operand.is_variable() or rule_node.value.operand.is_constant():
                    return None
                else:
                    result_data[rule_node.value.operand.variable] = copy.copy(self)
        else:
            if rule_node.value.operand.is_numeric_value() or rule_node.value.operand.is_constant():
                if self.value != rule_node.value:
                    return None
            else:
                result_data[rule_node.value.operand.variable] = copy.copy(self)

        return result_data

    def replace(self, rule):
        substitutions = self.unify(rule[0].root)

        def dfs(rule_node, parent, substitutions):
            if rule_node.value.is_operand():
                if rule_node.value.operand.is_variable() and not rule_node.value.operand.is_constant():
                    new_node = copy.copy(substitutions[rule_node.value.operand.variable])
                else:
                    new_node = copy.copy(rule_node)
                parent.add_child(new_node)
            else:
                new_node = Node(copy.copy(rule_node.value))
                parent.add_child(new_node)

                for i in range(len(rule_node.children)):
                    dfs(rule_node.children[i], new_node, substitutions)

        tmp_root = Node(value=None)
        dfs(rule[1].root, tmp_root, substitutions)

        new_root = tmp_root.children[0]
        self.value = new_root.value
        self.children = new_root.children

    def set_shape(self):
        # for child_node in self.children:
        #     child_node.set_shapes()

        for child_node in self.children:
            if child_node.value.shape is None:
                raise ValueError('Set shape error. Children shape is not defined')

        if self.value.is_operand():
            if self.value.operand.shape is None:
                raise ValueError('Operand without shape')
            self.value.shape = self.value.operand.shape
        else:
            if isinstance(self.value.operation, BinOp):
                if not self.children[0].value.shape == self.children[1].value.shape == ():
                    raise ValueError('Wrong shapes for number binary operation.')
                self.value.shape = self.children[0].value.shape

            elif isinstance(self.value.operation, UnOp):
                if self.children[0].value.shape != ():
                    raise ValueError('Wrong shapes for number unary operation.')
                self.value.shape = self.children[0].value.shape

            elif isinstance(self.value.operation, VectorBinOp):
                if self.value.operation == VectorBinOp.Add or self.value.operation == VectorBinOp.Sub:
                    if self.children[0].value.shape != self.children[1].value.shape or\
                       self.children[0].value.shape == ():
                        raise ValueError('Wrong vector addition/subtraction arguments')
                    self.value.shape = self.children[0].value.shape

                elif self.value.operation == VectorBinOp.MultByNum:
                    if not (self.children[0].value.shape != () and self.children[1].value.shape == () or
                            self.children[0].value.shape == () and self.children[1].value.shape != ()):
                        raise ValueError('Wrong vector multiplication by number error')

                    if self.children[0].value.shape == ():
                        self.value.shape = self.children[1].value.shape
                    else:
                        self.value.shape = self.children[0].value.shape

                elif self.value.operation == VectorBinOp.DotProd:
                    if not (self.children[0].value.shape[0] == self.children[1].value.shape[0] == 1):
                        raise ValueError('Wrong vector dot product arguments')

                    self.value.shape = ()

                elif self.value.operation == VectorBinOp.GetInd:
                    if not (self.children[0].value.shape != () and self.children[1].value.shape == ()):
                        raise ValueError('Wrong vector multiplication by number error')

                    if self.children[0].value.shape[0] == 1:
                        self.value.shape = self.children[0].children[0].children[0].value.shape
                    else:
                        self.value.shape = self.children[0].children[0].value.shape

                elif self.value.operation == VectorBinOp.Mult:
                    if self.children[0].value.shape[1] != self.children[1].value.shape[0] or\
                       self.children[0].value.shape == ():
                        raise ValueError('Wrong vector multiplication arguments')
                    self.value.shape = (self.children[0].value.shape[0], self.children[1].value.shape[1])

                elif self.value.operation == VectorBinOp.TensMult:
                    if self.children[0].value.shape == () or self.children[1].value.shape == ():
                        raise ValueError('Wrong tensor multiplication arguments')
                    self.value.shape = (self.children[0].value.shape[0]*self.children[1].value.shape[0],
                                        self.children[0].value.shape[1]*self.children[1].value.shape[1])

            elif isinstance(self.value.operation, VectorUnOp):
                if self.children[0].value.shape == ():
                    raise ValueError('Wrong vector unary minus argument')
                if self.value.operation == VectorUnOp.USub:
                    self.value.shape = self.children[0].value.shape

                elif self.value.operation == VectorUnOp.Transpose:
                    if self.children[0].value.shape == ():
                        raise ValueError('Wrong vector transpose argument')
                    self.value.shape = (self.children[0].value.shape[1], self.children[0].value.shape[0])

            elif isinstance(self.value.operation, VectorOp):
                elem_shapes = set()
                for child_node in self.children:
                    elem_shapes.add(child_node.value.shape)
                if None in elem_shapes:
                    raise ValueError('Vector element shape is not set')
                if len(elem_shapes) == 0:
                    raise ValueError('Empty vector')
                if len(elem_shapes) > 1:
                    raise ValueError('Vector element shapes are different')

                elem_shape = list(elem_shapes)[0]
                if not(len(elem_shape) == 0 or len(elem_shape) == 2):
                    raise ValueError('Wrong subvector shapes')
                if elem_shape == ():
                    self.value.shape = (1, len(self.children))
                else:
                    self.value.shape = (len(self.children), elem_shape[1])


class NodeValue:
    """
    NodeValue can be whether an operation or operand.
    """

    def __init__(self, operation=None, operand=None, shape=None):
        """

        :param operation:
        :type operation:
        :param operand:
        :type operand:
        """

        self.shape = shape
        if operation is not None:
            self.type = 0
            self.operation = operation
        else:
            self.type = 1
            self.operand = operand
            if self.shape is None:
                self.shape = self.operand.shape

    def __str__(self):
        if self.is_operand():
            return str(self.operand)
        else:
            return str(self.operation.name)

    def __eq__(self, other):
        return self.is_operation() and other.is_operation() and self.operation == other.operation or\
                self.is_operand() and other.is_operand() and self.operand == other.operand

    def __copy__(self):
        if self.is_operation():
            return NodeValue(operation=self.operation, shape=self.shape)
        else:
            return NodeValue(operand=copy.copy(self.operand), shape=self.shape)

    def is_operation(self):
        return self.type == 0

    def is_operand(self):
        return self.type == 1



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_empty_expression(self):
        input_string = ''
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_unary_op(self):
        input_string = '+1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_binary_op_sum(self):
        input_string = '1+1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_binary_op_sub(self):
        input_string = '1-1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_binary_op_mul(self):
        input_string = '1*1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_binary_op_div(self):
        input_string = '1/1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_binary_op_pow(self):
        input_string = '1**1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_two_ops(self):
        input_string = '1+1+1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_two_ops2(self):
        input_string = '1+1-1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))


    def test_three_ops(self):
        input_string = '1/1*1-1'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_order_ops1(self):
        input_string = '1*2/3*4'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_order_ops2(self):
        input_string = '1+2*3/4-5'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_order_pow(self):
        input_string = '1**2*3'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_parentheses(self):
        input_string = '(1+1)*(2+2)/(3-4)+(5*6)'
        self.assertEquals(Tree(expression=input_string), Tree(expression=input_string))

    def test_not_equal1(self):
        input_string1 = '1+1'
        input_string2 = '1+2'
        self.assertNotEquals(Tree(expression=input_string1), Tree(expression=input_string2))

    def test_not_equal2(self):
        input_string1 = '1+1'
        input_string2 = '1*1'
        self.assertNotEquals(Tree(expression=input_string1), Tree(expression=input_string2))

    def test_not_equal3(self):
        input_string1 = '+1'
        input_string2 = '1'
        self.assertNotEquals(Tree(expression=input_string1), Tree(expression=input_string2))

    def test_not_equal4(self):
        input_string1 = '1*(1+1)'
        input_string2 = '(1+1)*1'
        self.assertNotEquals(Tree(expression=input_string1), Tree(expression=input_string2))

if __name__ == '__main__':
    unittest.main()



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_binary_op_sum(self):
        input_string = '1+1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2')

    def test_binary_op_sub(self):
        input_string = '1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0')

    def test_binary_op_mul(self):
        input_string = '1*1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_binary_op_div(self):
        input_string = '1/1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1.0')

    def test_binary_op_pow(self):
        input_string = '1**1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_two_ops(self):
        input_string = '1+1+1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_two_ops2(self):
        input_string = '1+1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_three_ops(self):
        input_string = '1/1*1-1'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0.0')

    def test_order_ops1(self):
        input_string = '1*2/8*4'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1.0')

    def test_order_ops2(self):
        input_string = '1+2*3/4-5'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '-2.5')

    def test_order_pow(self):
        input_string = '1**2*3'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_parentheses(self):
        input_string = '(1+1)*(2+2)/(3-4)+(5*6)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '22.0')

    def test_with_args1(self):
        input_string = '2+3+x'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '5+x')

    def test_with_args2(self):
        input_string = 't-(3+x)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, 't-(3+x)')

    def test_with_args3(self):
        input_string = '-(1+a)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '-(1+a)')

    def test_with_args4(self):
        input_string = '(2+3+a)*(1+a)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(5+a)*(1+a)')

    def test_with_args5(self):
        input_string = '1/(3*4+b+c)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1/((12+b)+c)')

    # 2+x-2
    def test_with_args6(self):
        input_string = '2**(x-2)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2**(x-2)')

    def test_with_args7(self):
        input_string = '+(1+3+x)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '+(4+x)')

    def test_with_args8(self):
        input_string = '2*(x)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2*x')

    def test_with_args9(self):
        input_string = '(3*x)/3'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(3*x)/3')

    def test_with_args10(self):
        input_string = '(x+1)*(1-2)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '(x+1)*(-1)')

    def test_with_args11(self):
        input_string = '((a+1)*c+b*a+b)/(b+c)'
        tree = Tree(expression=input_string)
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '((((a+1)*c)+(b*a))+b)/(b+c)')
        
    def test_vector_binary_op_sum(self):
        input_string = '[1+1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '2')

    def test_vector_binary_op_sub(self):
        input_string = '[1-1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0')

    def test_vector_binary_op_mul(self):
        input_string = '[1*1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_vector_binary_op_div(self):
        input_string = '[1/1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1.0')

    def test_vector_binary_op_pow(self):
        input_string = '[1**1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_vector_two_ops(self):
        input_string = '[1+1+1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_vector_two_ops2(self):
        input_string = '[1+1-1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1')

    def test_vector_three_ops(self):
        input_string = '[1/1*1-1]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '0.0')

    def test_vector_order_ops1(self):
        input_string = '[1*2/8*4]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '1.0')

    def test_vector_order_ops2(self):
        input_string = '[1+2*3/4-5]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '-2.5')

    def test_vector_order_pow(self):
        input_string = '[1**2*3]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '3')

    def test_vector_parentheses(self):
        input_string = '[(1+1)*(2+2)/(3-4)+(5*6)]'
        tree = Tree(expression=input_string)
        
        tree.tree_calculate()
        output_string = tree.to_expression()
        self.assertEquals(output_string, '22.0')

if __name__ == '__main__':
    unittest.main()



import copy
from sqc.src.Constant import constants


class Operand:
    """
    Operand can be whether a numeric value or variable of given dimension.
    """
    def __init__(self, numeric_value=None, variable=None, shape=None):
        """
        :param numeric_value:   Number/vector/matrix
        :type numeric_value:    int, float
        :param variable:        Name of the variable or list/matrix of variable.
        :type variable:         string
        """

        self._is_constant = False
        self.shape = shape
        if numeric_value is not None:
            self.shape = ()
            self.type = 0
            self.numeric_value = numeric_value
        elif variable is not None:
            self.type = 1
            self.variable = variable
            if self.variable in constants.keys():
                self.shape = ()
                self._is_constant = True

    def __str__(self):
        if self.is_numeric_value():
            return str(self.numeric_value)
        elif self.is_variable():
            return str(self.variable)
        else:
            return '[]'

    def __eq__(self, other):
        if self.is_variable():
            return other.is_variable() and self.variable == other.variable
        elif self.is_numeric_value():
            return other.is_numeric_value() and self.numeric_value == other.numeric_value

    def __copy__(self):
        if self.is_numeric_value():
            return Operand(numeric_value=self.numeric_value)
        elif self.is_variable():
            return Operand(variable=self.variable, shape=self.shape)

    def is_numeric_value(self):
        return self.type == 0

    def is_variable(self):
        return self.type == 1

    def is_constant(self):
        return self._is_constant



from distutils.core import setup


setup(
    # Application name:
    name='symbolic-quantum-computation',

    # Version number (initial):
    version='0.2.8',

    # Application author details:
    author='Savelyev Nikita',
    author_email='savelyevno@gmail.com',

    # Packages
    packages=['sqc', 'sqc/src', 'sqc/test_package', 'sqc/test_package/tests'],
    entry_points={
        'console_scripts': [
            'sqc = sqc.__main__:main'
        ]
    },

    # Include additional files into the package
    include_package_data=True,

    # Details
    url='https://bitbucket.org/grpdev/symbolic-quantum-computation',

    #
    license='LICENSE.txt',
    readme='README.txt',
    description='symbolic-quantum-computation description',

    # long_description=open('README.txt').read(),

    # Dependent packages (distributions)
    install_requires=[
        'termcolor'
    ]
)



import unittest

from sqc.src.Tree import Tree


class Test(unittest.TestCase):
    def test_binary_op_sum(self):
        input_string = '1+1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_binary_op_sub(self):
        input_string = '1-1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_binary_op_mul(self):
        input_string = '1*1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_binary_op_div(self):
        input_string = '1/1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_binary_op_pow(self):
        input_string = '1**1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_two_ops(self):
        input_string = '1+1+1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_two_ops2(self):
        input_string = '1+1-1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_three_ops(self):
        input_string = '1/1*1-1'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_order_ops1(self):
        input_string = '1*2/8*4'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_order_ops2(self):
        input_string = '1+2*3/4-5'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_order_pow(self):
        input_string = '1**2*3'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_parentheses(self):
        input_string = '(1+1)*(2+2)/(3-4)+(5*6)'
        tree = Tree(expression=input_string)
        
        self.assertEquals(tree.root.value.shape, ())

    def test_with_args1(self):
        input_string = '2+3+x'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args2(self):
        input_string = 't-(3+x)'

        var_shape = {'x': (), 't': ()}

        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args3(self):
        input_string = '-(1+a)'

        var_shape = {'a': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args4(self):
        input_string = '(2+3+a)*(1+a)'

        var_shape = {'a': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args5(self):
        input_string = '1/(3*4+b+c)'

        var_shape = {'a': (), 'b': (), 'c': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args7(self):
        input_string = '+(1+3+x)'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args8(self):
        input_string = '2*(x)'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args9(self):
        input_string = '(3*x)/3'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args10(self):
        input_string = '(x+1)*(1-2)'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_with_args11(self):
        input_string = '((a+1)*c+b*a+b)/(b+c)'

        var_shape = {'a': (), 'b': (), 'c': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_uni_vector(self):
        input_string = '[1]'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_uni_vector2(self):
        input_string = '[[1]]'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_empty_double_vector(self):
        input_string = '[[2], [1]]'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (2, 1))

    def test_vector1(self):
        input_string = '[x]'

        var_shape = {'x': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, ())

    def test_vector1_1(self):
        input_string = '[x]'

        var_shape = {'x': (1, 1)}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, (1, 1))

    def test_vector2(self):
        input_string = '[x, y]'

        var_shape = {'x': (), 'y': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, (1, 2))

    def test_vector2_1(self):
        input_string = '[x, y]'

        var_shape = {'x': (1, 3), 'y': (1, 3)}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, (2, 3))

    def test_vector3(self):
        input_string = '[[1, x], [2, y]]'

        var_shape = {'x': (), 'y': ()}
        tree = Tree(expression=input_string, var_shapes=var_shape)

        self.assertEquals(tree.root.value.shape, (2, 2))

    def test_vector_ops_Add_2(self):
        input_string = 'vAdd([3, 4], [1, 6])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (1, 2))

    def test_vector_ops_Add_4(self):
        input_string = 'vAdd([[3, 2]], [[1, 5]])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (1, 2))

    def test_vector_ops_Add_5(self):
        input_string = 'vAdd([[3, 5], [1, 6], [3, 7]], [[3, 5], [1, 6], [3, 7]])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (3, 2))

    def test_vector_ops_MulNum_2(self):
        input_string = 'vMulNum([3, 4], -2)'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (1, 2))

    def test_vector_ops_MulNum_3(self):
        input_string = 'vMulNum(1, [[3], [1]])'
        tree = Tree(expression=input_string)
        self.assertEquals(tree.root.value.shape, (2, 1))

    def test_vector_ops_MulNum_4(self):
        input_string = 'vMulNum(-2, [[3, 2], [1, 5]])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (2, 2))

    def test_vector_ops_DP_1(self):
        input_string = 'vDP([3], [1])'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_DP_2(self):
        input_string = 'vDP([3, 4], [1, 6])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_vector_ops_DP_3(self):
        input_string = 'vDP([3, 4], [1, 6])-27'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_vector_ops_DP_4(self):
        input_string = 'vDP([3, 4], [1, 6])/10'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_vector_ops_Get_1(self):
        input_string = 'vGet([3], 1)'

        with self.assertRaises(Exception):
            tree = Tree(expression=input_string)

    def test_vector_ops_Get_2(self):
        input_string = 'vGet([3, 4], 2)'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, ())

    def test_vector_ops_Get_3(self):
        input_string = 'vGet([[3], [4]], 1)'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (1, 1))

    def test_vector_ops_Get_4(self):
        input_string = 'vGet([[3, 4], [1, 6]], 2)'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (1, 2))

    def test_vector_ops_USub_1(self):
        input_string = 'vUSub([[3], [4]])'
        tree = Tree(expression=input_string)

        self.assertEquals(tree.root.value.shape, (2, 1))

    def test_with_args1_err(self):
        input_string = '2+3+x'

        with self.assertRaises(Exception):
            var_shape = {'x': (1, 1)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args2_err(self):
        input_string = 't-(3+x)'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (2, 2), 't': (2, 2)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args3_err(self):
        input_string = '-(1+a)'
        
        with self.assertRaises(Exception):
            var_shape = {'a': (1, 2)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args5_err(self):
        input_string = '1/(3*4+b+c)'
        
        with self.assertRaises(Exception):
            var_shape = {'a': (), 'b': (), 'c': (1, 1)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args7_err(self):
        input_string = '+(1+3+x)'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (2,)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args8_err(self):
        input_string = '2*(x)'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 1)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args9_err(self):
        input_string = '(3*x)/3'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 3)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args10_err(self):
        input_string = '(x+1)*(1-2)'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 1)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_with_args11_err(self):
        input_string = '((a+1)*c+b*a+b)/(b+c)'
        
        with self.assertRaises(Exception):
            var_shape = {'a': (), 'b': (), 'c': (1, 2)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_vector2_err(self):
        input_string = '[x, y]'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 1), 'y': (1, 2)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_vector2_1_err(self):
        input_string = '[x, y]'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 3), 'y': (3, 3)}
            tree = Tree(expression=input_string, var_shapes=var_shape)

    def test_vector3_err(self):
        input_string = '[[1, x], [2, y]]'
        
        with self.assertRaises(Exception):
            var_shape = {'x': (1, 1), 'y': (1, 2)}
            tree = Tree(expression=input_string, var_shapes=var_shape)
            

if __name__ == '__main__':
    unittest.main()
